class InvalidMoleculeError(Exception):
    """Raised when the mol block string is not a valid mol block string."""


class InvalidReactionError(Exception):
    """Raised when the reaction block string is not a valid reaction block
    string."""
